import { ITestWaiter, WaiterName, ITestWaiterDebugInfo, Token } from './types';
/**
 * A class providing creation, registration and async waiting functionality.
 *
 * @public
 * @class TestWaiter<T>
 */
export default class TestWaiter<T = Token> implements ITestWaiter<T> {
    name: WaiterName;
    private nextToken;
    private isRegistered;
    items: Map<T, ITestWaiterDebugInfo>;
    /**
     * @public
     * @constructor
     * @param name {WaiterName} the name of the test waiter
     */
    constructor(name: WaiterName, nextToken?: () => T);
    /**
     * Will register the waiter, allowing it to be opted in to pausing async
     * operations until they're completed within your tests. You should invoke
     * it after instantiating your `TestWaiter` instance.
     *
     * **Note**, if you forget to register your waiter, it will be registered
     * for you on the first invocation of `beginAsync`.
     *
     * @private
     * @method register
     */
    private register;
    /**
     * Should be used to signal the beginning of an async operation that
     * is to be waited for. Invocation of this method should be paired with a subsequent
     * `endAsync` call to indicate to the waiter system that the async operation is completed.
     *
     * @public
     * @method beginAsync
     * @param item {T} The item to register for waiting
     * @param label {string} An optional label to identify the item
     */
    beginAsync(token?: T, label?: string): T;
    /**
     * Should be used to signal the end of an async operation. Invocation of this
     * method should be paired with a preceeding `beginAsync` call, which would indicate the
     * beginning of an async operation.
     *
     * @public
     * @method endAsync
     * @param item {T} The item to that was registered for waiting
     */
    endAsync(token: T): void;
    /**
     * Used to determine if the waiter system should still wait for async
     * operations to complete.
     *
     * @public
     * @method waitUntil
     * @returns {boolean}
     */
    waitUntil(): boolean;
    /**
     * Returns the `debugInfo` for each item tracking async operations in this waiter.
     *
     * @public
     * @method debugInfo
     * @returns {ITestWaiterDebugInfo}
     */
    debugInfo(): ITestWaiterDebugInfo[];
    /**
     * Resets the waiter state, clearing items tracking async operations in this waiter.
     *
     * @public
     * @method reset
     */
    reset(): void;
}
